@@ -1,11 +1,6 @@
 package HPCI;
 ### HPCI.pm ###################################################################
 
-### HISTORY ###################################################################
-# Version       Date        Developer         Comments
-# 0.30          2015-04-09  John Macdonald    preparing initial CPAN releaase
-# 0.31          2015-05-29  John Macdonald    initial CPAN release
-
 ### INCLUDES ##################################################################
 
 # safe Perl
@@ -13,25 +8,80 @@ use warnings;
 use strict;
 use Carp;
 use Module::Load;
+use Module::Load::Conditional qw(can_load);
+
+our @extra_roles;
+
+sub add_extra_role {
+	# next line is documentation
+	# my ($cluster, $level, $role) = @_;
+	shift; # get rid of HPCI class name
+	push @extra_roles, [ @_ ];
+}
+
+sub get_extra_roles {
+	my ($target_cluster, $target_level) = @_;
+	my @roles;
+	for my $role_bunch (@extra_roles) {
+		my ($cluster, $level, $roles) = @$role_bunch;
+		next unless $cluster eq 'ALL' || $cluster eq $target_cluster;
+		next unless $level eq $target_level;
+		push @roles, ref $roles ? @$roles : $roles;
+	}
+	return @roles;
+}
+
+my $default_attrs = {};
+
+sub add_default_attrs {
+ 	my $newhash = ref($_[0]) eq 'HASH' ? shift : { @_ };
+	_merge_hash( $default_attrs, $newhash );
+}
+
+sub _merge_hash {
+	my( $target, $new, $path ) = @_;
+	$path ||= [];
+	croak "not a hash when merging attribute hash{".join('}{',@$path)."}"
+	    unless ref($target) eq 'HASH' && ref($new) eq 'HASH';
+	while (my($k,$v) = each %$new) {
+		if (ref($v) eq 'HASH' || (exists $target->{$k} && ref($target->{$k}) eq 'HASH')) {
+			$target->{$k} //= {};
+			_merge_hash( $target->{$k}, $v, [ @$path, $k ] );
+		}
+		else {
+			$target->{$k} = $v;
+		}
+	}
+}
 
 sub group {
-    my $pkg = shift;
-    my $args =
-        scalar(@_) == 1 && ref($_[0]) eq 'HASH' ? shift
-        : scalar(@_) % 2 == 0                   ? { @_ }
-        : croak("HPCI->group() requires a hashref or a hash in list form");
-    my $cluster = $args->{cluster} || croak("HPCI->group() requires a cluster key in the argument hash");
-    $args->{log_file} //= 'group.log';
-    my $clmod = "HPCD::${cluster}::Group";
-    load $clmod;
-    return $clmod->new($args);
+	my $pkg = shift;
+	my $args =
+		scalar(@_) == 1 && ref($_[0]) eq 'HASH' ? shift
+		: scalar(@_) % 2 == 0                   ? { @_ }
+		: croak("HPCI->group() requires a hashref or a hash in list form");
+	# copy the default attributes as a start
+	my $use_args = {};
+	_merge_hash( {}, $default_attrs );
+	my $cluster = $args->{cluster} // $use_args->{cluster}
+		// croak("HPCI->group() requires a cluster key in the argument hash");
+
+	for my $arg_set ($use_args, $args) {
+		if (my $spec_args = delete $arg_set->{cluster_specific}) {
+			_merge_hash( $use_args, $spec_args->{$cluster} // {} );
+		}
+	}
+	_merge_hash( $use_args, $args );
+	my $clmod = "HPCD::${cluster}::Group";
+	load $clmod;
+	return $clmod->new($use_args);
 }
 
 
 sub _trigger_mkdir {
-    my $self = shift; # an object with a log
-    my $dir  = shift; # a Path::Class::Dir object
-    $self->info( "Created directory: $_" ) for $dir->mkpath;
+	my $self = shift; # an object with a log
+	my $dir  = shift; # a Path::Class::Dir object
+	$self->info( "Created directory: $_" ) for $dir->mkpath;
 }
 
 =head1 NAME
@@ -40,11 +90,15 @@ sub _trigger_mkdir {
 
 =head1 VERSION
 
-Version 0.31
+Version 0.39
 
 =cut
 
-our $VERSION = '0.31';
+our $VERSION = '0.39';
+
+our $LocalConfigFound;
+
+$LocalConfigFound = can_load( modules => { 'HPCI::LocalConfig' => undef });
 
 =head1 SYNOPSIS
 
@@ -74,6 +128,7 @@ our $VERSION = '0.31';
         pre_reqs => [ qw(analysis_A analysis_B analysis_C) ],
         dep => 'report'
     );
+
     my $status_info = $group->execute;
 
     my $exit_status = 0;
@@ -84,6 +139,8 @@ our $VERSION = '0.31';
         }
     }
 
+	exit(0); # all stages completed without error
+
 =head1 OVERVIEW
 
 HPCI (High Performance Computing Interface) provides an interface to
@@ -125,11 +182,19 @@ However, using such extensions makes it harder to move to a
 different cluster type.  So, actually making use of such extensions
 must be considered carefully.
 
-=head1 INTERFACE
+=head1 The life cycle of a B<group>
 
-The life cycle of running a group of commands on a cluster is as follows:
+A B<group> is the main mechanism for using HPCI.  It is an object that
+manages a group of computation steps (called B<stage>s), distributing them
+across the cluster and keeping track of various housekeeping details like
+when each stage can be run, checking for the result of each completed stage
+run, deciding whether a failure should cause a stage to be retried to to
+prevent other stages from being executed, and collecting the status for each
+stage.
 
-=over
+The life cycle of running a group of commands on a cluster is:
+
+=over 4
 
 =item create group
 
@@ -146,7 +211,7 @@ A B<stage> is created for each command that is to be executed on a
 separate node of the cluster.  This is created using the B<group>
 object's method B<stage>.
 
-=item define dependency ordering
+=item define dependency ordering between the stages
 
 An important reason for running a group of jobs on a cluster is the
 ability to use multiple computers to run portions of the computation
@@ -161,21 +226,37 @@ method B<add_deps>.
 
 Finally, the B<group> method B<execute> will run the entire set
 of stages.  It does not return until all stages have completed (or
-have been skipped).  Each stage will normally be run once.  However,
-it is possible for there to be retries under some recoverable
-failure conditions causing some stages to be run .multiple times.
-Alternatively, a hard failure of one stage can be a trigger for
+have been skipped).  Each stage will normally be run once, however
+it is possible for some stages to be retried under some
+failure conditions.
+A failure of one stage (after retry possibilities have been exhausted)
+can be a trigger for
 completely skipping the execution of other stages.  Each separate
-execution of a stage is managed with an internal object (called a
-job - but a user program won't see job objects directly).
+execution of a stage (original or retry) is managed with an internal object
+called a job - but a user program won't see job objects directly.
+
+As many stages as possible are run simultaneously.  This is limited by
+the specified dependencies, by cluster-specific driver limits, and by
+user-specified limits on concurrent execution.
 
 =back
 
-=head2 OUTPUT DIRECTORY LAYOUT
+The objects that calling code deals with directly are a group object to
+manage a group of stages, and a stage object for each separately run job.
+Internally, there are also job objects for each retry of a stage, and a
+log object for logging the execution process (alternately, the user can
+provide their own Log4Perl compatible log object for HPCI to use - this may be
+of use if you wish to merge logging of multiple groups and/or of other
+processing within your program together in a single log).
 
-There are a number of output files created during a group execution.
+There are also some facilities to provide local customization of the standard
+usage of HPCI (see "Local Customization" below).
 
-The default layout of these files is:
+=head1 Output Tree Layout
+
+There are a number of output files and directories created during a group execution.
+
+The default layout of these is:
 
   <base_dir>                     "."
     <group_dir>                  <base_dir>/<name>-<YYYYMMDD-hhmmss>
@@ -188,7 +269,7 @@ The default layout of these files is:
         final_retry              symlink to final <job_dir>
 
 Many of these files/directories can be re-assigned to different
-location using group or stage attributes - shown avove is the
+location using group or stage attributes - shown above is the
 default layout.  Commonly, you will specifically use the I<base_dir>
 attribute to choose a location other than the current directory for
 placing the tree; or else use the I<group_dir> attribute if you want
@@ -197,36 +278,42 @@ to choose a location that does not create a sub-directory for you.
 may end up with a mixture of old and new contents that are hard to
 figure out.)
 
-=head3 base_dir
+=over 4
+
+=item base_dir
 
 The top level of all the generated output.  It defaults to ".",
-but can be specified explicitly when the group is created.
+but can be specified explicitly when the group is created with
+the attribute B<base_dir>.
 
-=head3 group_dir
+=item group_dir
 
 By default, a new directory is created under B<base_dir>.  Its name
 is I<name>-I<YYYYMMDD>-I<hhmmss> - the name of the group along with
-a timestamp of when the execution started.
+a timestamp of when the execution started.  This can be over-ridden
+when the group is created by providing the group attribute B<group_dir>.
 
-=head3 log
+=item log
 
-The automatically provided log is written to the file I<name>.log
+The automatically provided log is written to the file I<"group.log">
 directly under I<group_dir>.  This logs information about the
-execution of the entire group of stages.
+execution of the entire group of stages.  See B<Logging Attributes
+of group object> below for ways of changing the default setting.
 
-=head3 stage_dir
+=item stage_dir
 
 Each stage creates a sub-directory beneath I<group_dir> with the
-same name as the stage.
+same name as the stage.  An alternate name can be used by providing
+the B<dir> attribute when the stage object is created.
 
-=head3 script_file
+=item script_file
 
 The script created to be executed on the cluster node.  This wraps
 the specified command with additional logic to pass on environment
 and config info, and to set output redirection.  It is called
 "script.sh" and placed in I<stage_dir>.
 
-=head3 job_dir
+=item job_dir
 
 A sub-directory is created under I<stage_dir> for each attempt to
 run the command.  Usually, there will only be a single attempt.
@@ -238,13 +325,13 @@ the entire I<stage_dir> directory would not even get created).
 These directories are simply named with the retry number ("0",
 "1", ...).
 
-=head3 stdout/stderr
+=item stdout/stderr
 
 Within each I<job_dir>, the files "stdout" and "stderr" collect
 the standard output and standard error output from that (re)try
 attempt to run the command.
 
-=head3 final_retry
+=item final_retry
 
 A symlink named "final_retry" is created within I<stage_dir> that
 points to the I<job_dir> of the final (re)try.  Since you often
@@ -252,12 +339,17 @@ don't care as much about the initial run tries as you do about the
 last one, this symlink provides a consistant access path to that
 final retry.
 
-=head2 HPCI "Class"
+=back
 
-You can pretend that B<HPCI> is a class with one class method
-named B<group>.
+=head1 HPCI "Class" Methods
 
-=head3 METHOD group
+You can pretend that B<HPCI> is a class with one primary class
+method named B<group>.
+
+There a few other class methods used for localization purposes, they
+are decribed below in "Local Customization".
+
+=head2 B<group> method
 
 The B<group> method creates and returns a group object, which
 you can treat like a B<HPCI::Group> object.  (In fact, it really
@@ -265,50 +357,80 @@ returns an object of class B<HPCD::I<cluster>::Group>, but if you
 ignore that fact then you can trivially have your program run on
 some other cluster type.)
 
-The group object is described further below.
+=head2 B<group> object
 
-=head2 Group Object
+The description of attributes and methods for the B<group> object given here describe
+the generic attributes and how they are treated for all cluster types.
+Individual cluster drivers can modify this behaviour and can provide
+additional attributes and methods for cluster-specific purposes.
 
-A group object is created with the HPCI class method B<group>
-(and not using a B<new> method directly.
-
-=head3 Cluster Attributes
+=head3 Cluster-Related Attributes of B<group> object
 
 The one necessary attribute is B<cluster>.  For some specific
 cluster types there may be additional attributes required for
 connecting to the cluster software (authentification, usage
 class info, etc.).
 
-=head4 cluster
+=over 4
 
-The cluster attribute specifies which type of cluster is to be used.
+=item cluster
+
+The B<cluster> attribute specifies which type of cluster is to be used.
 This is the only required attribute.  (Some cluster types may have
 additional attributes that are required for specifying connection
 and authentification info.)
 
-=head3 Basic Attributes
+=item cluster_specific
 
-=head4 name
+The attribute B<cluster_specific> is optional.  If provided, it should
+contain a hashref of hashrefs.  If the value specified for the I<cluster>
+attribute is present as a key in the B<cluster_specific>
+hash, the corresponding value will be used as a set of attribute values
+when the group is created.  Its elements will replace or augment any values
+for the same attribute name provided to the group method.  This will normally be
+used if the program can be dynamically configured for different cluster
+types, and there are different arg settings required for the different
+types of cluster.
 
-The name you give to a group is used for creating the directory
+=back
+
+=head3 Basic Attributes of B<group> object
+
+=over 4
+
+=item name
+
+The B<name> you give to a group is used for creating the directory
 where output is stored, and also in log messages.  A default name
 "default_group_name" is provided if you do not specific an explicit
 name.  Using the default name is adequate in simple programs which
 only create one group, but for more complicated programs giving
 separate names to each group is necessary to easily identify the
-output of each group.
+output of each group.  The value of B<name> may also be used by
+the cluster-specific driver to provide an identifier name (or the
+basis of one) to the underlying cluster, if it needs one.
 
-=head3 Directory Layout Attributes
+=item stage_defaults
 
-=head4 base_dir
+The attribute B<stage_defaults> is optional.  If provided, it should
+contain a hashref.  This hash will be used as default values for
+every stage created by this group.
+
+=back
+
+=head3 Directory Layout Attributes of B<group> object
+
+=over 4
+
+=item base_dir
 
 If none of the other directory layout attributes are used to
 over-ride this, this attribute specifies the directory in which
 all output directories and files will be created.  This is
 usually an existing directory; it defaults to the current
-directort ".".
+directory ".".
 
-=head4 group_dir
+=item group_dir
 
 This directory is usually created to contain the outputs of the
 group execution.  By default, it is directly under B<base_dir> with
@@ -321,14 +443,20 @@ should not be an existing directory containing previous results.
 the stage directories will over-write equivalently named directories
 and files that are created in this run, while leaving unchanged any
 that did not recur, so you'll have a mix of old and new contents.)
+The names of files and directories created under B<group_dir> are
+chosen to be consistent and easy to find automatically.
 
-=head3 Logging Attributes
+=back
+
+=head3 Logging Attributes of B<group> object
 
 An HPCI group logs its activities using a Log::Log4perl logger.
 The logger can either be provided by the caller, or else HPCI will
 create its own.
 
-=head4 log
+=over 4
+
+=item log
 
 This a Log::Log4perl::Logger object.  If it is provided as an
 attribute to the B<group> creation call, it will be used as it is,
@@ -339,39 +467,43 @@ object will be created using the attributes below to define where
 it is logged to.  This created logger will send all log entries to
 a file, as well as sending all info and higher log entries to stderr.
 
-=head4 log_path
+=item log_path
 
 If this attribute is provided (and the B<log> attribute is not
 provided) it will be used as the full pathname of a file where the
 log will be written.  If it is not provided, it will use the path
 B<log_dir>/B<log_file> by default.
 
-=head4 log_dir
+=item log_dir
 
 If neither B<log> or B<log_path> is provided, this attribute can
 be used to specify the directory where the log file is to be written.
 By default, it uses B<group_dir>.
 
-=head4 log_file
+=item log_file
 
 If neither B<log> or B<log_path> is provided, this attribute can
 be used to specify the file name to be written in the log directory.
-By default, it uses I<name>.log.
+By default, it uses the constant name "group.log".
 
-=head4 log_level
+=item log_level
 
 You can provide this attribute to change the default log level setting from "info" to any of I<debug info warn error fatal>.
 
-=head3 Operational Attributes
+=back
 
-=head4 max_concurrent
+=head3 Operational Attributes of B<group> object
+
+=over 4
+
+=item max_concurrent
 
 This attribute specifies the maximum number of stages that will
 be executing at one time.  The default setting of 0 allows as
 many stages as possible (all those that are not waiting for a
 pre-requisite stage to complete) to run at the same time.
 
-=head4 status
+=item status
 
 This attribute is set internally while stages are executed.
 It contains the final result status from each stage run that
@@ -385,84 +517,113 @@ a hash.  The key B<exit_status> contains the exit status of the run.
 If the stage was never run, B<exit_status> instead contains a text
 message listing the reason that it was skipped.
 
-=head3 Environment Passing Attributes
+=back
+
+=head3 Environment Passing Attributes of B<group> object
 
 You can set up a set of enviroment variables that will be provided to
 all stages.  (You can also set variables that are only for individual
-stages - if so, they will modfy any set you provide in the group.)
+stages - if so, they will modify any set you provide in the group.)
 
 See B<HPCI::Env> for a description of these.
 
-=head3 Methods
-
-=head4 stage
+=head3 Method B<stage> of B<group> object
 
 The method B<stage> is used to create a new stage object.
-Its attributes are described below.
+Its characteristics are described below.
 
-=head4 add_deps
+The B<group> object keeps track of all B<stage> objects created
+within that group so that they can all be managed properly when the
+B<execute> method is invoked.
+
+=head3 Method B<add_deps> of B<group> object
 
 The method B<add_deps> is used to specify pre-requisite/dependent
 relationships.  It takes either a hashref or a list containing
-two pairs.  One key must be either B<pre_req> or B<pre_reqs>,
-the other must be either B<dep> or B<deps>.
+pairs.  One of the keys must be either B<pre_req> or B<pre_reqs>,
+another must be either B<dep> or B<deps>.
 
-If the singular key is used (either B<pre_req> or B<dep>), then
-the corresponding value must be a scalar - either the string name
-of a stage object, or a reference to a stage object.
+The value for each of these keys can be either a scalar, or an arrayref
+of scalar values.  A scalar value can be either a B<stage> object
+or the name (string) of a stage object - and refer to a stage that has been
+created within this group.
 
-If the plural key is used (either B<pre_reqs> or B<deps>), then the
-corresponding value must be an arrayref containing a list of scalars,
-each one being either the string name of a stage or a reference to
-a stage object.
+HPCI will ensure that the stage or all of the stages specified for pre_req
+or pre_reqs have completed execution before any of the dep (or deps) stages
+is allowed to start executing.
 
 The plural forms are provided for convenience - often the output
 file from one preparation stage is required by many others, or the
 output from many processing stages is needed by a stage that merges
-results into a summary report.
+results into a summary report.  Rather than having to loop over the
+pre_reqs and deps and calling B<add_dep> individually for every
+individual dependency, a single call will handle the entire combination.
+
+While it is recommended for code readability that you use the singular
+form (B<dep> or B<pre_req>) is you are providing a single stage, and the
+plural form (B<deps> or B<pre_reqs>) if you are providing a list of
+stages, either can be used.
+
+The B<add_dep> method can be called multiple times.  HPCI will
+accumlate the dependencies appropriately.
 
 It is an error to provide a sequence of dependencies that form
-a cycle.
+a cycle in which a stage directly or indirectly has itself as a
+pre-requisite.  (Such a stage could never run.  HPCI will detect
+when all remaining stages are blocked by pre-requisites and abort,
+but that might be after numerous stages have already been executed.)
 
-=head4 execute
+=head3 METHOD execute of B<group> object
 
 The B<execute> method is the final goal of building the group.
-It schedules the execution of individual stages.  It allows for
-dependencies before running a stage.  It provides for re-running
+It schedules the execution of individual stages.  It waits for
+pre-requisites before running a stage.  It provides for re-running
 a stage if a soft failure has occurred that allows a retry.  If a
-failure that cannot be retried occurrs, it skips scheduling dependent
-stages, or stops scheduling all new stages.
+failure that cannot be retried occurs, it can skip scheduling dependent
+stages, or even stop scheduling all new stages.
 
 =head2 Stage Object
 
 =head3 Attributes
 
-=head4 name
+=over 4
 
-The B<name> attribute must be provided for stages.  It is a string.
-Each stage within a group must have a unique name.  There is no
-default value provided.
+=item name
 
-=head4 command
+A unique B<name> attribute must be provided for stages.  It is a string.
+There is no default value provided.
+
+=item command
 
 The B<command> attribute must be provided before the group is
-executed.  It can either be provided as a string attribute in the
-B<stage> method call that creates the stage, or using the one of
+executed.  It can either be provided as a string attribute when the
+stage is created, or by using the one of
 the command-setting methods provided by the stage class.
 
 See B<HPCI::Stage> for more details about the command setting
 methods.
 
-=head4 dir
+=item dir
 
 The B<dir> attribute is optional.  It specifies the direcory
 in which files related to the stage are placed.  By default,
 it is I<group_dir>/I<stage_name>.  You will usually not need to
 change this.
 
-=head4 resources_required
+=item cluster
 
-=head4 retry_resources_required
+The B<cluster> attribute is automatically passed on fro mthe B<group>
+to each B<stage>.  You are not likely to need this.
+
+=item group
+
+The B<group> that created a stage is automatically passed on (as a weak
+reference) to the stage.  You are not likely to need to use this attribute
+in user code.
+
+=item resources_required
+
+=item retry_resources_required
 
 The B<resources_required> and B<retry_resources_required> are used to
 define resources that will be required by the stage when it executes.
@@ -481,14 +642,28 @@ from this list.
 
 See B<HPCI::Stage> for more details about resources.
 
-=head4 failure_action ('abort_group', 'abort_deps'*, or 'ignore')
+=item force_retries
+
+This attribute specifies an integer number of time to retry the
+stage before comcluding that it has actually failed.  You might use
+this if your cluster has some nodes that work differently from
+others and a stage might fail on one type of node but succeed on
+another.
+
+These retries are done after any cluster-specific retry mechanisms
+have been used.
+
+The default value for this attribute is 0 (zero), giving no forced
+retries unless you specifically ask for them.
+
+=item failure_action ('abort_group', 'abort_deps'*, or 'ignore')
 
 Specifies the action to take if this stage fails (terminates with
 a non-zero status).
 
 There are three string values that it can have:
 
-=over
+=over 4
 
 =item - abort_deps (default)
 
@@ -510,20 +685,43 @@ no longer blocked.
 
 =back
 
+=item abort_group_on_failure abort_deps_on_failure ignore_failure
+
 As an alternative to providing a value to the failute_action attribute
 when you create a stage, you can instead provide one of the pseudo-attributes
 'abort_group_on_failure', 'abort_deps_on_failure', or 'ignore_failure' with
 a true value to specify 'abort_group', 'abort_deps', or 'ignore' respectively.
 
+=item state
+
+The B<state> is mostly an internal attribute but after the group has
+finished execution you can use this to check whether the stage was
+run successfully. After execution, B<state> will either be 'pass" or
+'fail'.
+
+=item Environment passing attributes
+
+You can set up a set of environment variables that will be provided to
+this stage.  It will use set defined for the group as a basis (if such a set was
+defined for the group), but that set can be changed for individual stages
+or you can have no group default and only provide a set to specific stages
+as needed.  See B<HPCI::Env> for further details.
+
+=back
+
 =head3 Methods
 
 =head4 command creation
 
+There are a number of helper methods to assist in building different
+types of commands to be provided for the B<command> attribute.
 See B<HPCI::Stage> for details.
 
-=head3 Env
+=head1 Local Configuration
 
-See B<HPCI::Env> for details.
+TODO: write this section
+ - describe the HPCI::LocalConfig module
+ - describe the mechanism for adding extra roles to group, stage, etc.
 
 =head1 Additional
 
@@ -567,7 +765,7 @@ changes will only be made for critical reasons.
 The reason for separate distribution of cluster-specific HPCD
 packages are fairly obvious:
 
-=over
+=over 4
 
 =item -
 
@@ -586,7 +784,7 @@ download the driver modules that they actually need.
 
 =head1 SEE ALSO
 
-=over
+=over 4
 
 =item HPCI::Group
 
