@@ -11,6 +11,7 @@ use File::ShareDir;
 use File::Path qw(make_path);
 use Time::HiRes qw(usleep gettimeofday);
 use Try::Tiny;
+use DateTime;
 
 use MooseX::Role::Parameterized;
 use MooseX::Types::Path::Class qw(Dir File);
@@ -25,21 +26,21 @@ role {
 	my $theDriver = $p->theDriver;
 	my $StageClass = "${theDriver}::Stage";
 
-	method _trigger_mkdir => sub {
-		goto \&HPCI::_trigger_mkdir;
-	};
-
 	has '_stage_class' => (
-		is => 'ro',
-		isa => 'Str',
+		is       => 'ro',
+		isa      => 'Str',
 		init_arg => undef,
-		default => $StageClass,
+		default  => $StageClass,
 	);
 
 	sub name;         # tell Logger we provide a name attribute
 	sub group_dir;    # tell Logger we provide a group_dir attribute
 	sub _unique_name; # tell Logger we provide a group_dir attribute
 
+	# do nothing, but let roles write before/after/around wrappers
+	method BUILD => sub {
+	};
+
 	with qw(
 		HPCI::Logger
 		HPCI::Env
@@ -52,15 +53,16 @@ role {
 	#                        - now generated dynamically
 	#                        - allowing for easier control by plugins
 
-	sub wait_for_completed_stage {
+	method wait_for_completed_stage => sub {
 		my $self         = shift;
 		my $finished_job = $self->next_finished_job;
 		my $name         = $finished_job->stage->name;
 		my $stage        = $self->_stages->{$name};
-		$self->info("stage($name) completed!");
 		$stage->_analyse_completion_state($finished_job);
+		my $state = $stage->state;
+		$self->info("stage($name) completion state: $state");
 		return $stage;
-	}
+	};
 
 =head1 NAME
 
@@ -192,10 +194,10 @@ when the group was created (e.g. EXAMPLEGROUP-YYMMDD-HHMMSS).
 			my $self = shift;
 			my $subdir = $self->_unique_name;
 			my $target = $self->base_dir->subdir($subdir);
-			$self->_trigger_mkdir( $target );
+			HPCI::_trigger_mkdir( $self, $target );
 			return $target;
 		},
-		trigger => \&_trigger_mkdir,
+		trigger => \&HPCI::_trigger_mkdir,
 	);
 
 =head2 max_concurrent (optional)
@@ -213,6 +215,20 @@ limits of its own).
 		default  => 0,
 	);
 
+=head2 stage_defaults
+
+This attribute can be given a hash reference containing values that
+will be passed to every stage created.
+
+=cut
+
+	has 'stage_defaults' => (
+		is       => 'ro',
+		isa      => 'HashRef',
+		lazy     => 1,
+		default  => sub { {} },
+	);
+
 =head2 status (provided internally)
 
 After the execute method has been called, this attribute contains the
@@ -311,8 +327,8 @@ object is consistant with the actual group object and that you don't
 have to change code in multiple places if you switch to using a
 different cluster type for the group.  (If you want to mix stages
 for multiple cluster types within your program, you should either
-create two groups that execute independently, or er else create a
-stage that itself run HPCI and manages the stages for the second
+create two groups that execute independently, or else create a
+stage that itself creates a group and manages the stages for the second
 type of cluster.)
 
 The name parameter is required and must be unique - two stages
@@ -327,12 +343,23 @@ reference to the object.)
 
 	requires qw(submit_stage);
 
-	sub stage {
+	method stage => sub {
 		my $self = shift;
 		$self->_croak("Cannot define new stages after execution has started!")
-		    if $self->_execution_started;
-		my $stage = $self->_stage_class->new(
-			@_,
+			if $self->_execution_started;
+		my $use_args = {};
+		HPCI::_merge_hash( $use_args, $self->stage_defaults );
+		my $args     = { @_ };
+		my $cluster  = $self->cluster;
+		for my $arg_set ($use_args, $args) {
+			if (my $spec_args = delete $arg_set->{cluster_specific}) {
+				HPCI::_merge_hash( $use_args, $spec_args->{$cluster} // {} );
+			}
+		}
+		HPCI::_merge_hash( $use_args, $args );
+
+		my $stage    = $self->_stage_class->new(
+			%$use_args,
 			cluster => $self->cluster,
 			group   => $self,
 		);
@@ -406,7 +433,7 @@ object or the name of the stage.
 
 =cut
 
-	sub add_deps {
+	method add_deps => sub {
 		my $self = shift;
 		my %defargs = (
 			dep => undef,
@@ -414,9 +441,16 @@ object or the name of the stage.
 			pre_req => undef,
 			pre_reqs => [],
 		);
-		my %args     = ( %defargs, @_);
+		my %args     = ( %defargs, scalar(@_) == 1 ? %{ $_[0] } : @_ );
 		my $badargs  = join ' ', grep { ! exists $defargs{$_} } keys %args;
 		$self->_croak( "Unknown arg($badargs) provided to add_deps" ) if $badargs;
+		#		for my $arg ( qw(dep pre_req) ) {
+		#			my $args = $arg.'s';
+		#			$self->_croak( "Arg ($args) must be an array ref" )
+		#				unless ref($args{$args}) eq 'ARRAY';
+		#			$self->_croak( "Arg ($arg) may not be an array ref" )
+		#				if ref($args{$arg}) eq 'ARRAY';
+		#		}
 		my @deps     = _build_stage_list($args{dep}, $args{deps});
 		my @pre_reqs = _build_stage_list($args{pre_req}, $args{pre_reqs});
 		for my $dep (@deps) {
@@ -424,7 +458,7 @@ object or the name of the stage.
 				$self->_add_dep( $pre_req, $dep );
 			}
 		}
-	}
+	};
 
 	sub _build_stage_list {
 		return
@@ -434,14 +468,14 @@ object or the name of the stage.
 			@_
 	}
 
-	sub _add_dep { # little cost/no damage if called again with same dep and pre_req
+	method _add_dep => sub { # little cost/no damage if called again with same dep and pre_req
 		my ( $self, $pre_req, $dep ) = @_;
 		return if $self->_deps->{$pre_req}{$dep};
 		$self->_deps->{$pre_req}{$dep}     = 1;
 		$self->_pre_reqs->{$dep}{$pre_req} = 1;
 		$self->_blocked->{$dep}            = 1;
 		delete $self->_ready->{$dep};
-	}
+	};
 
 =head2 $group->execute
 
@@ -451,7 +485,7 @@ other stage or the attempt is aborted).
 
 =cut
 
-	sub execute {
+	method execute => sub {
 		my $self = shift;
 		my $catcher = sub {
 			my $sig = shift;
@@ -464,15 +498,14 @@ other stage or the attempt is aborted).
 		$self->_execution_started(1);
 		local $SIG{CHLD} = $SIG{CHLD};
 		$self->child_catcher;
-		$self->info(
-			"Starting execution with ",
-			$self->_stage_cnt,
-			" stages defined."
-		);
+		my $cnt = $self->_stage_cnt;
+		my $s   = $cnt == 1 ? '' : 's';
+		$self->info( "Starting execution with $cnt stage$s defined.");
 		while ($self->_finished_cnt < $self->_stage_cnt) {
 			$self->_submit_ready;
 			unless (keys %{ $self->_submitted }) {
 				my $blocked = join( ', ', keys %{ $self->_blocked } );
+				$self->_discard_deps( "Deadlock" );
 				$self->_croak("Deadlock - cannot submit any more jobs, remaining jobs ($blocked) are blocked, there must be a dependency loop.")
 			}
 			$self->_await_one_job;
@@ -482,10 +515,33 @@ other stage or the attempt is aborted).
 		while (my ($name, $stage) = each %{ $self->_stages }) {
 			$ret_stat->{$name} = $stage->get_run_stats;
 		}
-		$self->info( "All done!  Stage status:", Dumper($ret_stat) );
+		# $self->info( "All done!  Stage status:", Dumper($ret_stat) );
+		$self->info( "All done!  Stage status:" );
+		for my $jobname (sort keys %$ret_stat) {
+			my $jobs = $ret_stat->{$jobname};
+			for my $jobnum (0..$#$jobs) {
+				my $job = $jobs->[$jobnum];
+				my $jobnumdesc =
+					$jobnum != $#$jobs ? "Retried run $jobnum"
+					: $jobnum          ? "Final run after $jobnum retries"
+					:                    "Only run - no retries";
+				for my $key (sort keys %$job) {
+					my $val = $job->{$key};
+					if (defined $jobname) {
+						$self->info( "  $jobname" );
+						undef $jobname;
+					}
+					if (defined $jobnumdesc) {
+						$self->info( "    $jobnumdesc" );
+						undef $jobnumdesc;
+					}
+					$self->info( sprintf "      %-20s => %s", $key, $val );
+				}
+			}
+		}
 		$self->_set_status($ret_stat);
 		return $ret_stat;
-	}
+	};
 
 	# sub wait_for_all_jobs {
 	# 	# TODO: Never called, so the html report is never written!
@@ -526,7 +582,7 @@ other stage or the attempt is aborted).
 	#     Kill off any executing stages, issue an error message, and quit.
 	#     Make sure that the quit happens even if the rest of the cleanup fails
 
-	sub _croak {
+	method _croak => sub {
 		my $self = shift;
 		my $msg  = shift;
 		my $fatalfailed;
@@ -537,7 +593,7 @@ other stage or the attempt is aborted).
 			++$fatalfailed;
 		};
 		try {
-			$self->kill_stages;
+			$self->kill_stages if $self->_finished_cnt < $self->_stage_cnt;
 		}
 		catch {
 			try {
@@ -546,24 +602,24 @@ other stage or the attempt is aborted).
 			}
 		};
 		croak $msg;
-	}
+	};
 
-	sub _croak_extra {
+	method _croak_extra => sub {
 		return;
-	}
+	};
 
 	# _submit_ready
 	#   - submit all stages that are ready to be run
 	#   - that will be stages which have no pre_reqs, or for which all pre_reqs
 	#     have completed successfully
 
-	sub _submit_ready {
+	method _submit_ready => sub {
 		my $self              = shift;
 		my $min_submit_window = 500_000;    # 0.5 second
 		my $min_delay         = 1_000;      # 1 millisecond is not worth delaying over
 		$self->info("Searching for unblocked stages to submit.\n");
 		for my $name ( keys %{ $self->_ready } ) {
-			return if $self->max_concurrent && $self->concurrent >= $self->_running_cnt;
+			return if $self->max_concurrent && $self->max_concurrent <= $self->_running_cnt;
 			delete $self->_ready->{$name};
 			my $stage = $self->_stages->{$name};
 			$stage->assert_command_filled;
@@ -572,8 +628,8 @@ other stage or the attempt is aborted).
 			my ( $now_secs, $now_microsecs ) = gettimeofday;
 			if ($now_secs < $self->_sub_secs + 2) {
 				my $delta_microsecs =
-				    ( $now_secs - $self->_sub_secs ) * 1_000_000 +
-				    ( $now_microsecs - $self->_sub_microsecs );
+					( $now_secs - $self->_sub_secs ) * 1_000_000 +
+					( $now_microsecs - $self->_sub_microsecs );
 				$delta_microsecs = $min_submit_window - $delta_microsecs;
 				if ($delta_microsecs > $min_delay) {
 					$self->info(
@@ -594,32 +650,32 @@ other stage or the attempt is aborted).
 			$self->info("Submitting stage ($name).\n");
 			$self->submit_stage( $stage );
 		}
-	}
+	};
 
 	# _unblock_deps
 	#   - after a stage has finished successfully, any stages that are
 	#     dependent upon it can be (possibly partially) unblocked
 	#   - if this is the final block for a dependent, it is marked ready to run
 
-	sub _unblock_deps {
+	method _unblock_deps => sub {
 		my $self    = shift;
 		my $pre_req = shift;
 		for my $dep ( keys %{ $self->_deps->{$pre_req} } ) {
 			delete $self->_pre_reqs->{$dep}{$pre_req};
 			$self->info( "removing dep: $pre_req -> $dep, remaining deps are( "
-				  . join( ' ', sort keys %{ $self->_pre_reqs->{$dep} } )
-				  . " )" );
+				. join( ' ', sort keys %{ $self->_pre_reqs->{$dep} } )
+				. " )" );
 			unless (keys %{ $self->_pre_reqs->{$dep} }) {
-				delete $self->_blocked->{$dep};
-				$self->_ready->{$dep} = 1;
+				$self->_ready->{$dep} = 1
+					if delete $self->_blocked->{$dep};
 			}
 		}
-	}
+	};
 
 	# _blocked_deps_list
 	#   - get all deps recursively
 
-	sub _blocked_deps_list {
+	method _blocked_deps_list => sub {
 		my $self   = shift;
 		my $target = shift;
 		my $seen   = shift || {};
@@ -630,7 +686,7 @@ other stage or the attempt is aborted).
 			$self->_blocked_deps_list( $dep, $seen );
 		}
 		return $seen;
-	}
+	};
 
 	# _discard_deps
 	#   - after a stage has failed. the action to be taken is determined by the
@@ -644,9 +700,9 @@ other stage or the attempt is aborted).
 	#     - ignore
 	#       - continue with all other stages
 
-	sub _discard_deps {
+	method _discard_deps => sub {
 		my $self    = shift;
-		my $failed  = shift;
+		my $failure = shift;
 		my $pre_req = shift;
 		my @discards = sort keys %{
 			$pre_req
@@ -655,21 +711,21 @@ other stage or the attempt is aborted).
 		};
 		for my $stage (@discards) {
 			$self->error(
-				"Skipping stage ($stage) because of failure of $failed");
+				"Skipping stage ($stage) because of $failure");
 			$self->_finished_cnt( $self->_finished_cnt + 1 );
 			delete $self->_blocked->{$stage};
 			$self->_failed->{$stage} = 1;
 			$self->_stages->{$stage}->_set_failure_info(
-				"Skipped because of failure of previous stage $failed");
+				"Skipped because of $failure");
 		}
-	}
+	};
 
 	# _await_one_job
 	#   - wait until a running stage completes
 	#     - if it failed: remove any stages that depend upon it
 	#       (or all remaining jobs if abort_group_on_failure was set)
 
-	sub _await_one_job {
+	method _await_one_job => sub {
 		my $self  = shift;
 		my $stage = $self->wait_for_fully_completed_stage;
 		my $name  = $stage->name;
@@ -680,17 +736,19 @@ other stage or the attempt is aborted).
 		if ($stage->is_pass) {
 			$self->info("Successful completion of stage ($name).\n");
 			$self->_completed->{$name} = 1;
-			$self->_unblock_deps($name);
 		}
 		else {
 			# job failed and cannot be resubmitted
 			$self->error( "Failed stage ($name).\n" );
 			$self->_failed->{$name} = 1;
 			my $action = $stage->failure_action;
-			$self->_discard_deps( $name, ( $action eq 'abort_group' ? () : $name ) )
-			    unless $action eq 'ignore';
+			$self->_discard_deps(
+				"failure of stage $name",
+				( $action eq 'abort_group' ? () : $name ) )
+				unless $action eq 'ignore';
 		}
-	}
+		$self->_unblock_deps($name);
+	};
 
 	# wait for a stage to complete
 	#   - if it failed, determine whether it can be retried
@@ -700,24 +758,33 @@ other stage or the attempt is aborted).
 	# All that is required is to actually resubnit it for execution
 	#
 	# return the first completed stage that does not need to be restarted
-	sub wait_for_fully_completed_stage {
+	method wait_for_fully_completed_stage => sub {
 		my $self = shift;
 		while (1) {
 			my $stage = $self->wait_for_completed_stage;
+			if ($stage->is_fail && $stage->_forced_retries < $stage->force_retries) {
+				$stage->_set_state('retry');
+				$stage->_forced_retries($stage->_forced_retries+1);
+				$self->info( "Using forced retry: ",
+				   	$stage->_forced_retries,
+				   	" of ",
+				   	$stage->force_retries,
+				);
+			}
 			return $stage unless $stage->is_retry;
 			$self->retry($stage);
 		}
-	}
+	};
 
 	# this method can be over-ridden if cluster-specific code needs to take
 	# any action different from the submit done initially that can't be
 	# handled already in the wait_for_completed_stage method
-	sub retry {
+	method retry => sub {
 		my $self  = shift;
 		my $stage = shift;
 		$self->warn( "Retrying stage(".$stage->name.")");
 		$self->submit_stage($stage);
-	}
+	};
 
 };
 
@@ -729,7 +796,7 @@ John Macdonald         - Boutros Lab
 
 =head1 ACKNOWLEDGEMENTS
 
-Paul Boutros, Phd, PI - Boutros Lab
+Paul Boutros, Phd, PI - Boutros http://www.omgubuntu.co.uk/2016/03/vineyard-wine-configuration-tool-linuxLab
 
 The Ontario Institute for Cancer Research
 
