@@ -24,15 +24,15 @@ around BUILDARGS => sub {
 
     # support historical boolean parms for a while
     if ( delete $parms{abort_group_on_failure} ) {
-        $parms{on_failure} = 'abort_group';
+        $parms{failure_action} = 'abort_group';
     }
 
     if ( delete $parms{abort_deps_on_failure} ) {
-        $parms{on_failure} = 'abort_deps';
+        $parms{failure_action} = 'abort_deps';
     }
 
     if ( delete $parms{ignore_failure} ) {
-        $parms{on_failure} = 'ignore';
+        $parms{failure_action} = 'ignore';
     }
 
     return $class->$orig(%parms);
@@ -501,6 +501,30 @@ sub is_complete {
     return $self->is_pass || $self->is_fail;
 }
 
+=head2 force_retries
+
+If a cluster has a tendency to fail randomly, it may be desired to
+retry a stage that fails rather than simply treating it as an unrecoverable
+failure.
+
+This attribute specifies the number of times that the stage should be retried
+until it passes.  The default is to not retry (unless the underlying driver
+retries for a specific reason).
+
+=cut
+
+has 'force_retries' => (
+	is        => 'ro',
+	isa       => 'Int',
+	default   => 0
+);
+
+has '_forced_retries' => (
+	is        => 'rw',
+	isa       => 'Int',
+    init_arg  => undef,
+	default   => 0
+);
 
 has '_failure_info' => (
     is        => 'ro',
@@ -546,7 +570,7 @@ sub get_run_stats {
 
 =head2 failure_action ('abort_group', 'abort_deps'*, or 'ignore')
 
-Specifies the action to take if this stage fails (terminates with
+Specifies the action to take if (the final retry of) this stage fails (terminates with
 aa non-zero status).  There are three string values that it can have:
 
 =over
